/**
 * Code Binaire pour Machine Virtuelle à Pile.
 * Les flottants sont des double qui prennent la place de 2 ints.
 * Les adresses sont des int.
 * CALL empile l'adresse de retour et la vauler du framePointer dans la pile.
 */
public class CBaP {

    
    
    pc = 0;
    fp = 0;
    stack;
    program;
    const SizeofDouble = 2;

    _debug = false;
    
    CBaP(program, debug, stackSize) {
	this.program = program;
	this._debug = debug;
	this.stack = new Pile(stackSize);
    }

    dumpInstruction(ii, out) {
	let i = ii;
	switch(program.get_int(i)) {                                               
	case AssembleMVaP.ADD:    out.print("ADD   ");            	break;
	case AssembleMVaP.SUB:    out.print("SUB   ");            	break;
	case AssembleMVaP.MUL:    out.print("MUL   ");            	break;
	case AssembleMVaP.DIV:    out.print("DIV   ");            	break;
	case AssembleMVaP.INF:    out.print("INF   ");            	break;
	case AssembleMVaP.INFEQ:  out.print("INFEQ ");          	break;
	case AssembleMVaP.SUP:    out.print("SUP   ");            	break;
	case AssembleMVaP.SUPEQ:  out.print("SUPEQ ");      	break;
	case AssembleMVaP.EQUAL:  out.print("EQUAL ");      	break;
	case AssembleMVaP.NEQ:  out.print("NEQ   ");              	break;
								   	
	case AssembleMVaP.FADD:  out.print("FADD  ");            	break;     	
	case AssembleMVaP.FSUB:  out.print("FSUB  ");            	break; 		
	case AssembleMVaP.FMUL:  out.print("FMUL  ");            	break; 		
	case AssembleMVaP.FDIV:  out.print("FDIV  ");        	        break;   	
	case AssembleMVaP.FINF:  out.print("FINF  ");         	        break; 		 		
	case AssembleMVaP.FINFEQ:  out.print("FINFEQ");        	break; 		 	
	case AssembleMVaP.FSUP:  out.print("FSUP  ");            	break; 		 		
	case AssembleMVaP.FSUPEQ:  out.print("FSUPEQ");      	break; 		 	
	case AssembleMVaP.FEQUAL:  out.print("FEQUAL");      	break; 		 	
	case AssembleMVaP.FNEQ:  out.print("FNEQ  ");            	break; 		 		
	case AssembleMVaP.ITOF:  out.print("ITOF  ");        	        break; 		 		
	case AssembleMVaP.FTOI:  out.print("FTOI  ");        	        break; 		 		
								   	
	case AssembleMVaP.RETURN: out.print("RETURN");      	break;
	case AssembleMVaP.POP:    out.print("POP   ");            	break;      
	case AssembleMVaP.READ:   out.print("READ  ");           	break;      
	case AssembleMVaP.WRITE:  out.print("WRITE ");          	break;      
	case AssembleMVaP.WRITEF:  out.print("WRITEF");      	break;      
	case AssembleMVaP.PADD:    out.print("PADD  ");          	break;    
	case AssembleMVaP.PUSHGP:  out.print("PUSHGP");     	break;    
	case AssembleMVaP.PUSHFP:  out.print("PUSHFP");      	break;    
	case AssembleMVaP.DUP:    out.print("DUP   ");            	break;      

	case AssembleMVaP.PUSHI:  out.print("PUSHI ");  out.format("%7d ", program.get_int(++i));  break;
	case AssembleMVaP.PUSHG:  out.print("PUSHG ");  out.format("%7d ", program.get_int(++i));  break;
	case AssembleMVaP.STOREG: out.print("STOREG"); out.format("%7d ", program.get_int(++i));  break;
	case AssembleMVaP.PUSHL:  out.print("PUSHL ");  out.format("%7d ", program.get_int(++i));  break;
	case AssembleMVaP.STOREL: out.print("STOREL"); out.format("%7d ", program.get_int(++i));  break;
	case AssembleMVaP.PUSHR:    out.print("PUSHR "); out.format("%7d ", program.get_int(++i));  break;    
	case AssembleMVaP.STORER:  out.print("STORER"); out.format("%7d ", program.get_int(++i));  break;    
	case AssembleMVaP.FREE:    out.print("FREE  "); out.format("%7d ", program.get_int(++i));  break;	
	case AssembleMVaP.ALLOC:  out.print("ALLOC "); out.format("%7d ", program.get_int(++i));  break;

	case AssembleMVaP.PUSHF:  out.print("PUSHF ");  ++i; out.format("%7.3f ", program.get_double(++i));  break;

	case AssembleMVaP.JUMP:   out.print("JUMP  ");   out.format("%7d ", program.get_int(++i));  break;
	case AssembleMVaP.JUMPF:  out.print("JUMPF ");  out.format("%7d ", program.get_int(++i));  break;
	case AssembleMVaP.JUMPI: out.print("JUMPI "); out.format("%7d ", program.get_int(++i));  break;
	case AssembleMVaP.CALL:   out.print("CALL  ");   out.format("%7d ", program.get_int(++i));  break;
	case AssembleMVaP.HALT:   out.print("HALT  ");            	break;      
	default: System.err.println("Code inconnu "+program.get_int(i)+" ligne "+i);
	}
	if (i == ii) out.print("        ");
	return i;
    }
       
    public void dumpProgram(PrintStream out) {
	// out.println(program.toString(true));
	// out.println(program.toString());
	out.println(" Adr |  Instruction"); 
	out.println("-----+---------------"); 
        for(int i = 0; i < program.getSize(); i++) {
            out.format("%4d | ", i); 
	    i = dumpInstruction(i, out);
            out.println();
        }
    }

    private void _call(int label) {
	push(pc+1); 
	push(fp); 
	fp = size(); 
	pc = label; 
    }

    private void _return() {
	while(size() > fp)
	    pop();
	fp = pop();
	pc = pop();
    }

    private BufferedReader input = null;
    private int _read() {
	try {
	    if (input == null)
		input = new BufferedReader(new InputStreamReader(System.in));
	    return Integer.parseInt(input.readLine());
	} catch (Exception e) {
	    System.err.println(e);
	    System.exit(1);
	    return 0;
	}
    }

    public boolean execute() {
	if (_debug) {
	    dumpProgram(System.out);
	    System.err.println();
	    System.err.println("  pc |               |    fp   pile");
	    System.err.println("====================================================");
	}
	pc = 0;
	while (true) {
	    if (_debug) {
		System.err.format("%4d | ", pc);
		dumpInstruction(pc, System.err);
		System.err.format("|  %4d %s\n", fp, stack.toString());
	    }
	    try {
		int p1, p2; // utile pour être sûr de l'ordre des pop() !
		double d1, d2;
		switch(program.get_int(pc++)) {                                               
		case AssembleMVaP.ADD:    p1 = pop(); p2 = pop(); push(p2 +  p1);  	break;
		case AssembleMVaP.SUB:    p1 = pop(); p2 = pop(); push(p2 -  p1);  	break;
		case AssembleMVaP.MUL:    p1 = pop(); p2 = pop(); push(p2 *  p1);  	break;
		case AssembleMVaP.DIV:    p1 = pop(); p2 = pop(); push(p2 /  p1);   	break;
		case AssembleMVaP.INF:    p1 = pop(); p2 = pop(); push(p2 <  p1 ? 1 : 0);	break;
		case AssembleMVaP.INFEQ:  p1 = pop(); p2 = pop(); push(p2 <= p1 ? 1 : 0);	break;
		case AssembleMVaP.SUP:    p1 = pop(); p2 = pop(); push(p2 >  p1 ? 1 : 0);  	break;
		case AssembleMVaP.SUPEQ:  p1 = pop(); p2 = pop(); push(p2 >= p1 ? 1 : 0);  	break;
		case AssembleMVaP.EQUAL:  p1 = pop(); p2 = pop(); push(p2 == p1 ? 1 : 0); 	break;
		case AssembleMVaP.NEQ:    p1 = pop(); p2 = pop(); push(p2 != p1 ? 1 : 0); 	break;

		case AssembleMVaP.FADD:   d1 = pop_double(); d2 = pop_double(); push(d2 +  d1); break;
		case AssembleMVaP.FSUB:   d1 = pop_double(); d2 = pop_double(); push(d2 -  d1); break;
		case AssembleMVaP.FMUL:   d1 = pop_double(); d2 = pop_double(); push(d2 *  d1); break;
		case AssembleMVaP.FDIV:   d1 = pop_double(); d2 = pop_double(); push(d2 /  d1); break;
		case AssembleMVaP.FINF:   d1 = pop_double(); d2 = pop_double(); push(d2 <  d1 ? 1 : 0);	break;
		case AssembleMVaP.FINFEQ: d1 = pop_double(); d2 = pop_double(); push(d2 <= d1 ? 1 : 0);	break;
		case AssembleMVaP.FSUP:   d1 = pop_double(); d2 = pop_double(); push(d2 >  d1 ? 1 : 0);  	break;
		case AssembleMVaP.FSUPEQ: d1 = pop_double(); d2 = pop_double(); push(d2 >= d1 ? 1 : 0);  	break;
		case AssembleMVaP.FEQUAL: d1 = pop_double(); d2 = pop_double(); push(d2 == d1 ? 1 : 0);  	break;
		case AssembleMVaP.FNEQ:   d1 = pop_double(); d2 = pop_double(); push(d2 != d1 ? 1 : 0); 	break;

		case AssembleMVaP.FTOI:   push((int)pop_double()); 		break;
		case AssembleMVaP.ITOF:   push((double)pop()); 			break;

		case AssembleMVaP.RETURN: _return();                    	break; 
		case AssembleMVaP.POP:    pop();                        	break;      
		case AssembleMVaP.READ:   push(_read());                	break;      
		case AssembleMVaP.WRITE:  System.out.format("%7d\n", peek());   break;      
		case AssembleMVaP.WRITEF: System.out.format("%7.3f\n", peek_double()); break;      
		case AssembleMVaP.PADD:   p1 = pop(); p2 = pop(); push(p2 +  p1); break;
		case AssembleMVaP.PUSHGP: push(0);                              break;  
		case AssembleMVaP.PUSHFP: push(fp);                             break;  
		case AssembleMVaP.DUP:    push(peek());                         break;      
		    
		case AssembleMVaP.PUSHI:  push(program.get_int(pc++));          break;
		case AssembleMVaP.PUSHG:  push(get(program.get_int(pc++)));   	break;
		case AssembleMVaP.STOREG: set(program.get_int(pc++), pop());	break;
		    // ajouter -2 si adresse négative ?
		case AssembleMVaP.PUSHL:  push(get(fp+program.get_int(pc++))); 	break;
		case AssembleMVaP.STOREL:set(fp+program.get_int(pc++), pop());	break;
		case AssembleMVaP.PUSHR: push(get(pop()+program.get_int(pc++)));break;
		case AssembleMVaP.STORER:p1 = pop(); p2 = pop(); set(p2+program.get_int(pc++), p1); break;
		case AssembleMVaP.FREE: 	  doPop(program.get_int(pc++)); break;
		case AssembleMVaP.ALLOC:  doPush(program.get_int(pc++));        break;

		case AssembleMVaP.PUSHF:  push(program.get_int(pc++)); push(program.get_int(pc++)); break;

		case AssembleMVaP.JUMP:   pc = program.get_int(pc);		break;
		case AssembleMVaP.JUMPF:  if (pop() == 0) pc = program.get_int(pc); else pc++; break;
		case AssembleMVaP.JUMPI:  pc = program.get_int(pc)+pop();	break;
		case AssembleMVaP.CALL:   _call(program.get_int(pc));		break;
		case AssembleMVaP.HALT:   return true;               
		default: System.err.println("Code inconnu "+program.get_int(pc)+" ligne "+pc); return false;
		}
	    } catch (ArrayIndexOutOfBoundsException e) {
		System.err.println(stack.getException());
		return false;
	    }
        }
    }

    void push(int i) { stack.push(i); }
    int pop() { return stack.pop_int(); }
    void push(double d) { stack.push(d); }
    double pop_double() { return stack.pop_double(); } 
    int peek() { return stack.peek_int(); }
    double peek_double() { return stack.peek_double(); }
    int size() { return stack.getSize(); }
    void set(int index, int i) { stack.set(index, i); }
    int get(int index) { return stack.get_int(index); }

    void doPop(int nb) { for(; nb > 0; nb--) stack.pop_int(); }
    void doPush(int nb) { for(; nb > 0; nb--) stack.push(0); }
}
