// MVàP

let debugLexer = false;
let debug = true;

/** token lexicaux */
const T_ID = 0;
const T_INT = 1;
const T_FLOAT = 2;
const T_SPACE = 3;
const T_NL = 4;
const T_ERROR = 5;
const T_EOF = 6;

T_Text = [ "T_ID", "T_INT", "T_FLOAT", "T_SPACE", "T_NL", "T_ERROR", "T_EOF" ];
    
/** token grammaticaux */
const INSTR1 = 0;
const INSTR2 = 1;
const INT = 2;
const INSTR2F = 3;
const FLOAT = 4;
const LABEL = 5;
const JUMP = 6;
const NEWLINE = 7;

class MVaP {
    
    codes = [ // pour l'interprète de MVàP
	[ "ADD", INSTR1, () => { let p1 = this.stack.pop(); let p2 = this.stack.pop(); this.stack.push(p2+p1); } ],
	[ "SUB", INSTR1, () => { let p1 = this.stack.pop(); let p2 = this.stack.pop(); this.stack.push(p2-p1); } ],
	[ "MUL", INSTR1, () => { let p1 = this.stack.pop(); let p2 = this.stack.pop(); this.stack.push(p2*p1); } ],
	[ "DIV", INSTR1, () => { let p1 = this.stack.pop(); let p2 = this.stack.pop(); this.stack.push(Math.floor(p2/p1)); } ],
	[ "INF", INSTR1, () => { let p1 = this.stack.pop(); let p2 = this.stack.pop(); this.stack.push(p2<p1?1:0); } ],
	[ "INFEQ", INSTR1, () => { let p1 = this.stack.pop(); let p2 = this.stack.pop(); this.stack.push(p2<=p1?1:0); } ],
	[ "SUP", INSTR1, () => { let p1 = this.stack.pop(); let p2 = this.stack.pop(); this.stack.push(p2>p1?1:0); } ],
	[ "SUPEQ", INSTR1, () => { let p1 = this.stack.pop(); let p2 = this.stack.pop(); this.stack.push(p2>=p1?1:0); } ],
	[ "EQUAL", INSTR1, () => { let p1 = this.stack.pop(); let p2 = this.stack.pop(); this.stack.push(p2==p1?1:0); } ],
	[ "NEQ", INSTR1, () => { let p1 = this.stack.pop(); let p2 = this.stack.pop(); this.stack.push(p2!=p1?1:0); } ],
	[ "FADD", INSTR1, () => { let d1 = this.stack.popFloat64(); let d2 = this.stack.popFloat64(); this.stack.pushFloat64(d2+d1); } ],
	[ "FSUB", INSTR1, () => { let d1 = this.stack.popFloat64(); let d2 = this.stack.popFloat64(); this.stack.pushFloat64(d2-d1); } ],
	[ "FMUL", INSTR1, () => { let d1 = this.stack.popFloat64(); let d2 = this.stack.popFloat64(); this.stack.pushFloat64(d2*d1); } ],
	[ "FDIV", INSTR1, () => { let d1 = this.stack.popFloat64(); let d2 = this.stack.popFloat64(); this.stack.pushFloat64(d2/d1); } ],
	[ "FINF", INSTR1, () => { let d1 = this.stack.popFloat64(); let d2 = this.stack.popFloat64(); this.stack.push(d2<d1?1:0); } ],
	[ "FINFEQ", INSTR1, () => { let d1 = this.stack.popFloat64(); let d2 = this.stack.popFloat64(); this.stack.push(d2<=d1?1:0); } ],
	[ "FSUP", INSTR1, () => { let d1 = this.stack.popFloat64(); let d2 = this.stack.popFloat64(); this.stack.push(d2>d1?1:0); } ],
	[ "FSUPEQ", INSTR1, () => { let d1 = this.stack.popFloat64(); let d2 = this.stack.popFloat64(); this.stack.push(d2>=d1?1:0); } ],
	[ "FEQUAL", INSTR1, () => { let d1 = this.stack.popFloat64(); let d2 = this.stack.popFloat64(); this.stack.push(d2==d1?1:0); } ],
	[ "FNEQ", INSTR1, () => { let d1 = this.stack.popFloat64(); let d2 = this.stack.popFloat64(); this.stack.push(d2!=d1?1:0); } ],
	[ "ITOF", INSTR1, () => { let p1 = this.stack.pop(); this.trace("ITOF "+p1); this.stack.pushFloat64(p1); } ],
	[ "FTOI", INSTR1, () => { this.stack.push(Math.round(this.stack.popFloat64())); } ],
	[ "RETURN", INSTR1, () => { while(this.stack.getSize() > this.fp) this.stack.pop();
				    this.fp = this.stack.pop(); this.pc = this.stack.pop(); } ],
	[ "POP", INSTR1, () => { this.stack.pop(); } ],
	[ "READ", INSTR1, () => { this.stack.push(0); } ], // on lit toujours 0 (idée : prendre le contenu d'un input de type text spécial
	[ "WRITE", INSTR1, () => { this.trace(this.stack.peek()+"\n"); } ],
	[ "WRITEF", INSTR1, () => { this.trace(this.stack.peekFloat64()+"\n"); } ],
	[ "PADD", INSTR1, () => { let p1 = this.stack.pop(); let p2 = this.stack.pop(); this.stack.push(p2+p1); } ],
	[ "PUSHGP", INSTR1, () => { this.stack.push(0); } ],
	[ "PUSHFP", INSTR1, () => { this.stack.push(this.fp); } ],
	[ "DUP", INSTR1, () => { this.stack.push(this.stack.peek()); } ],
	
	[ "PUSHI", INSTR2, () => { this.stack.push(this.program.get(this.pc++)); } ],
	[ "PUSHG", INSTR2, () => { this.stack.push(this.stack.get(this.program.get(this.pc++)));} ],
	[ "STOREG", INSTR2, () => { this.stack.set(this.program.get(this.pc++), this.stack.pop())} ],
	[ "PUSHL", INSTR2, () => { this.stack.push(this.stack.get(this.fp+this.program.get(this.pc++))); } ],
	[ "STOREL", INSTR2, () => { this.stack.set(this.fp+this.program.get(this.pc++), this.stack.pop()); } ],
	[ "PUSHR", INSTR2, () => { this.stack.push(this.stack.get(this.stack.pop()+this.program.get(this.pc++))); } ],
	[ "STORER", INSTR2, () => { let p1 = this.stack.pop(); let p2 = this.stack.pop(); this.stack.set(p2+this.program.get(this.pc++)); } ],
	[ "FREE", INSTR2, () => { this.stack.doPop(this.program.get(this.pc++)); } ],
	[ "ALLOC", INSTR2, () => { this.stack.doPush(this.program.get(this.pc++)); } ],
	[ "PUSHF", INSTR2F, () => { this.stack.pushFloat64(this.program.getFloat64(this.pc)); this.pc += 2; } ],
	[ "CALL", JUMP, () => {	this.stack.push(this.pc+1); this.stack.push(this.fp); 
				this.fp = this.stack.getSize();
				this.pc = this.program.get(this.pc); } ],
	[ "JUMP", JUMP, () => { this.pc = this.program.get(this.pc); } ],
	[ "JUMPF", JUMP, () => { if (this.stack.pop()==0) this.pc = this.program.get(this.pc);
				 else this.pc++; } ],
	[ "JUMPI", JUMP, () => { this.pc = this.program.get(this.pc)+this.stack.pop();} ],

	[ "HALT", INSTR1, () => { return true; } ],

	[ "LABEL", LABEL, /* erreur interne */ ]
    ];
    /*** Ajouter instr : READF WRITEF POPF */
    
    tokens = []; // indexe le tableau codes sur son premier élément

    text = "";

    // <Lexer>
    // see LexerMVaP.png for the DFA diagram
    state = "Start";
    text; // le texte à analyser
    index = 0; // index dans la chaîne text
    lineNumber = 1;
    
    transitions = [
	{ state: "Start", next: "Start", test: (c) => { return c == "\n"; }, token: T_NL },
	{ state: "Start", next: "Start", test: (c) => { return c.trim() == ""; }, token: false },
	{ state: "Start", next: "Id", test: (c) => { return "A" <= c && c <= "Z"; }, token: false },
	{ state: "Id", next: "Id", test: (c) => { return "A" <= c && c <= "Z"; }, token: false},
	{ state: "Id", next: "Start", test: (c) => { return c.trim() == ""; }, token: T_ID},
	{ state: "Start", next: "Int", test: (c) => { return ("0" <= c && c <= "9")||c=="-"||c=="+"; }, token: false },
	{ state: "Int", next: "Int", test: (c) => { return "0" <= c && c <= "9"; }, token: false },
	{ state: "Int", next: "Start", test: (c) => { return c.trim() == ""; }, token: T_INT },
	{ state: "Int", next: "Decimal", test: (c) => { return c == "."; }, token: false },
	{ state: "Int", next: "E", test: (c) => { return c == "E"; }, token: false },
	{ state: "Decimal", next: "Decimal", test: (c) => { return "0" <= c && c <= "9"; }, token: false },
	{ state: "Decimal", next: "E", test: (c) => { return c == "E"; }, token: false },
	{ state: "Decimal", next: "Start", test: (c) => { return c.trim() == ""; }, token: T_FLOAT },
	{ state: "E", next: "Exponent", test: (c) => { return ("0"<=c && c<="9")||c=="-"||c=="+"; }, token: false },
	{ state: "Exponent", next: "Exponent", test: (c) => { return "0" <= c && c <= "9"; }, token: false },
	{ state: "Exponent", next: "Start", test: (c) => { return c.trim() == ""; }, token: T_FLOAT }
	// else: error
    ];

    initLexer() {
	this.index = 0;
	this.lineNumber = 1
	console.log("*****************\n*  C'est parti  *\n*****************");
    }
    
    // Donne la transition correspondante à l'état courant et au caractère entrant
    transition(c) {
	for (let v of this.transitions.values()) {
	    // on aurait pu indexer les transitions sur state au lieu de tout énumérer
	    if (this.state == v.state && v.test(c)) {
		return v;
	    }
	}
	return T_ERROR;
    }

    yytext = ""; // contient le texte du jeton syntaxique (nommé comme dans lex/yacc)

 
    nextToken() {
	let c;
	let trans;
	while (true) {
	    c = this.text[this.index];
	    if (c === undefined) 
		return T_EOF;
	    c = c.toUpperCase();
	    trans = this.transition(c);
	    if (trans == T_ERROR) { this.yytext += c; return T_ERROR; }
	    if (this.state != trans.next && this.state == "Start") this.yytext ="";

	    this.state = trans.next;
	    this.index++;

	    if (Number.isInteger(trans.token)) {
		break;
	    }
	    this.yytext += c;
	}

	// séparateur significatif
	if (c == "\n") {
	    if (trans.token != T_NL) {
		this.index--; // ne pas manger la fin de ligne non prise en compte
		// une autre façon de dire est que l'on n'incrémente pas l'index sur le renvoi de token
		// sauf en cas de token constitué de séparateur
	    } else 
		this.lineNumber++; // on en profite pour compter la ligne
	}
	// console.log(" line " + this.lineNumber + " " + trans.token + " '" + yytex + "' i="+i);
	return trans.token;
    }
    // </Lexer>

    // <Analyse sémantique et assemblage>

    /** Grammaire des instructions
       instr 
       : INSTR1 T_NL   
       | INSTR2  T_INT T_NL
       | JUMP  T_INT T_NL
       | INSTR2F T_FLOAT T_NL
       | LABEL T_NL
       | T_NL
       ;
    */

    program;// Pile contenant le programme compilé
    adressesLabels; // adresse des étiquettes pour l'assemblage

    analyse() {

	this.initLexer();
	
	let token; // le jeton lexical courant
	let state = NEWLINE; // état de l'analyseur grammatical (ici le précédent syntagme grammatical)
	this.program = new Pile();
	this.adressesLabels = [];
	
	while (true) {
	    this.yytext = "";
	    token = this.nextToken();
	    if (debugLexer)
		console.log("** nextToken line "+this.lineNumber+" token "+token+" "+T_Text[token]+" '" +this.yytext+"' index "+this.index+" state "+state);

	    switch(token) {
	    case T_EOF:
		return true;
	    case T_ID:
		let k = this.tokens[this.yytext];
		if (k === undefined || state != NEWLINE) return false;
		state = this.codes[k][1];
		if (state != LABEL) { // pas d'instruction LABEL dans le code
		    this.program.push(k);
		}
		break;
	    case T_INT:
		if (state == LABEL) { // on mémorise juste l'adresse pour la passe d'assemblage
		    this.adressesLabels[Number(this.yytext)] = this.program.getSize(); // ± 1 who knows?
		} else if(state == INSTR2 || state == JUMP) {
		    this.program.push(Number(this.yytext));
		} else return false;
		state = INT;
		break;
	    case T_FLOAT:
		if (state != INSTR2F) return false;
		state = FLOAT;
		this.program.pushFloat64(Number(this.yytext));
		break;
	    case T_NL:
		if (state == INSTR2 || state == JUMP || state == INSTR2F) return false;
		state = NEWLINE;
		break;
	    case T_ERROR:
		return false;
	    default:
		alert("Erreur interne prévenir la maintenance d'urgence");
		return false;
	    }
	}
	console.log("Programme correct syntaxiquement");
	return true;
    }

    assemble() {
	// change les numéros d'étiquettes par les adresses des instructions pointées.
	console.log("Label "+this.adressesLabels);
	console.log(this.dumpProgram());

	for (let i = 0; i < this.program.getSize(); i++) {
	    let k = this.program.get(i);
	    // Attention il faut compter si par malchance un entier = code d'un saut
	    switch (this.codes[k][1]) {
	    case INSTR1:
		break;
	    case INSTR2:
		i++; // on saute l'entier qui suit
		break;
	    case INSTR2F:
		i += 2; // on saute le flottant qui suit
		break;
	    case JUMP: // met à jour l'adresse du saut
		i++;
		console.log("--- set "+i);
		this.program.set(i, this.adressesLabels[this.program.get(i)]);
		break;
	    default:
		alert("Erreur assemblage");
	    }
	}
	// console.log(this.dumpProgram());
    }
    // </Analyse sémantique et assemblage>


    dumpInstruction(i, result) {
  	let code = this.codes[this.program.get(i)];
	result += code[0].padEnd(6, ' ');
	switch (code[1]) {
	case INSTR2:
	case JUMP:
	    result += this.program.get(++i).toString().padStart(7, ' ');
	    break;
	case INSTR2F:
	    result += this.program.getFloat64(++i).toString().padStart(7, ' ');
	    i++;
	    break;
	default:
	    result += "       ";
	}

	return [i, result];
    }
    
    dumpProgram() {
	let result = "";
	result += this.program + "\n";

	result += " Adr |  Instruction\n-----+---------------\n";
	for(let i = 0; i < this.program.getSize(); i++) {
	    result += i.toString().padStart(4, " ")+" | ";
	    [i, result] = this.dumpInstruction(i, result);
            result += "\n";
        }
	return result;
    }
    
    constructor() {
	// init tokens
	for (let k of this.codes.keys())
	    this.tokens[this.codes[k][0]] = k;
    }
    
    // charge et compile le code
    load(text) {
	this.text = text;
	// on insère une fin de ligne si le texte n'en a pas à sa fin
	if (text[text.length-1] != "\n") this.text = text+"\n";
	if (! this.analyse()) {
	    alert("Erreur ligne "+this.lineNumber+" sur '" +this.yytext+"'");		
	    return;
	}
	this.assemble();
    }

    pc = 0;
    fp = 0;
    stack;
    
    runcode(maxSteps, d) {
	debug = d;
	//for (let i = 0; i < this.codes.length; i++) console.log(i+ " "+ this.codes[i][0]);
	this.pc = 0;
	this.fp = 0;
	this.stack = new Pile();

	try {
	    if (debug)
		this.trace("Programme\n"+this.dumpProgram() + "\nTraces exécution\n"
			   + "  pc |  Instruction  |  fp   pile\n"
			   + "=====+===============+=============================\n");
	    
	    for (let i= 0; i < maxSteps; i++) // while (true)
	    {
		if (this.pc < 0 || this.pc >= this.program.getSize())
		    break;
		
		if (debug) {
		    let s = this.pc.toString().padStart(4, " ")+" | ";
		    let junk;
		    [junk, s] = this.dumpInstruction(this.pc, s);
		    s += " |"+this.fp.toString().padStart(4,' ')+'  '+this.stack.toString()+"\n";
		    this.trace(s);
		}
		
		if (this.codes[this.program.get(this.pc++)][2]() === true)
		    break;
	    }
	} catch(e) {
	    this.trace(e);
	    console.error(e);
	}
    }
	
    trace(s) {
	traces.innerText += s;
	console.log(s);
    }

}






