// Class Pile
// Auteur : Jacques Madelaine

// Une Pile est essentiellement une pile d'entiers ("int32")
// Elle peut aussi empiler et dépiler des flottants définis sur 64 bits.
// L'index est sur les entiers.

// Rappel : en JavaScript les Number sont tous stockés comme des flottants 64 bits.
// Il faut donc voir cette Pile comme un objet pédagogique
// et non une implémentation optimale en JavaScipt. 

"use strict";

class Pile {
    buffer; // ArrayBuffer de bytes
    pile;   // On va utiliser 4 bytes pour les entiers et 8 pour les flottants
    haut = 0; // index de sommet de pile en bytes 

    // size : taille en bytes de la pile
    constructor(size) {
	if (size === undefined)
	    size = 1000;
	
	this.buffer = new ArrayBuffer(size);
	this.pile = new DataView(this.buffer);
    } 

    push(i) {
	if (! Number.isInteger(i))
	    throw new TypeError("Try to push a non integer");
	this.pile.setInt32(this.haut, i);
	this.haut += 4;
    }

    doPush(nb) { for(; nb > 0; nb--) this.push(0); }
    
    pop() {
	this.haut -= 4;
	return this.pile.getInt32(this.haut);
    }

    doPop(nb) { for(; nb > 0; nb--) this.pop(); }
    
    peek() {
	return this.pile.getInt32(this.haut-4);
    }

    set(index, i) { 
	if (! Number.isInteger(i))
	    throw new TypeError("Try to set a non integer");
	this.pile.setInt32(index*4, i);
    }
   
    get(index) { 
	return this.pile.getInt32(index*4);
    }
    

    pushFloat64(f) {
	this.pile.setFloat64(this.haut, f);
	this.haut += 8;
    }

    popFloat64() {
	this.haut -= 8;
	return this.pile.getFloat64(this.haut);
    }

    peekFloat64() {
	return this.pile.getFloat64(this.haut-8);
    }

    setFloat64(index, f) { 
	this.pile.setFloat64(index*4, f);
    }
   
    getFloat64(index) { 
	return this.pile.getFloat64(index*4);
    }

    getMaxSize() {
	return this.buffer.byteLength/2;
    }

    getSize() {
	return this.haut/4;
    }
    
    toString() {
	let out = "[ ";
	for (let i = 0; i < this.getSize(); i++)
	    out += this.get(i) + ", ";
	return out + " ] " + this.getSize();
    }

}


/*

let p = new Pile();
try {
    let d = 12.34;
    let i = 0x7FFFFFFF;
    let j = 0;
    p.push(i);
    j = p.peek();
    console.log(i + " =?= " + j);

    let N = 5;
    for(let n = 0; n < N; n++) {
        p.push(n);
	console.log("push "+n+" " + n);
    }

    console.log("get(2) " + p.get(2));
    console.log(p.toString());
		       
    for(let n = N; n > 0; n--) {
        let m = p.pop();
    //        if (2*n != m)
        console.log("push "+ n + " pop " + m);
    }
    p.pop();
    
    p.push(40);
    p.pushFloat64(d);
    p.pushFloat64(13.34);
    p.push(i);
    console.log(p.toString());

    for (let n = 0; n<6;n++)
	console.log("p.getFloat64("+n+") = "+p.getFloat64(n));

    p.pop();
    p.popFloat64();
    let e = p.popFloat64();
    console.log(d + " =?= " + e);

    
    d = -1;
    p.pushFloat64(d);
    e = p.popFloat64();
    console.log(d + " =?= " + e);
    //
    //    for (i = 0 ; i < 1000000; i++)
    //     	p.push(i);
} catch(e) {
    console.error(e);
}
*/


